@@ -24,7 +24,9 @@
 
 setClass("big.matrix", representation(address='externalptr'))
 
-big.matrix = function(nrow, ncol, type='integer', init=0, dimnames=NULL)
+big.matrix = function(nrow, ncol, type='integer', init=NULL, dimnames=NULL,
+                      separated=FALSE, shared=FALSE, backingfile=NULL, 
+											backingpath=NULL, preserve=TRUE)
 {
   # It would be nice if init could be a vector or matrix.
   if (nrow < 1 | ncol < 1)
@@ -39,26 +41,69 @@ big.matrix = function(nrow, ncol, type='integer', init=0, dimnames=NULL)
     typeVal=2
   if (type == 'char') 
     typeVal=1
-
   if (is.null(typeVal)) stop('invalid type')
-  address = .Call('CCreateMatrix', nrow=as.double(nrow), ncol=as.double(ncol),
-                  init=as.double(init), type=as.integer(typeVal))
-  x = new("big.matrix", address=address)
-  if (is.null(x))
-    stop("Error encountered when creating instance of type big.matrix")
-  dimnames(x) = dimnames
-  return(x)
+  address=NULL
+  if (is.null(backingfile) & !shared)
+  {
+    address = .Call('CCreateMatrix', as.double(nrow), as.double(ncol),
+      as.double(init), as.integer(typeVal), as.logical(separated))
+    if (is.null(address))
+    {
+      stop(paste("Error: Memory could not be allocated for instance of",
+				"type big.matrix", sep=' '))
+    }
+    x = new("big.matrix", address=address)
+    if (is.null(x))
+      stop("Error encountered when creating instance of type big.matrix")
+    dimnames(x) = dimnames
+    return(x)
+  }
+  else
+  {
+#		if (!is.null(backingpath) &&
+#			substr(backingfile, nchar(backingfile), nchar(backingfile)) != '/' ||
+#			substr(backingfile, nchar(backingfile)-1, nchar(backingfile)) != '\\' )
+#		{
+#			backingfile = paste(backingfile, '/', sep='')
+#		}
+		print("backing file is")
+		print(backingfile)
+    return(shared.big.matrix(nrow=nrow, ncol=ncol, type=type, init=init, 
+			dimnames=dimnames, separated=separated, backingfile=backingfile, 
+			backingpath=backingpath, preserve=preserve))
+  }
 }
 
 is.big.matrix <- function(x) return(class(x) == "big.matrix")
 
-as.big.matrix <- function(x, type = 'integer')
+as.big.matrix <- function(x, type = NULL, separated = FALSE,
+  shared=FALSE, backingfile=NULL, backingpath=NULL, preserve=TRUE)
 {
-  if (type=="integer" | type=="double" | type=="short" | type=="char") {
-    y <- big.matrix(nrow(x), ncol(x), type, 0)
+  if (is.vector(x)) {
+    x <- matrix(x, length(x), 1)
+    warning("Coercing vector to a single-column matrix.")
+  }
+  if (!is.matrix(x)) 
+		stop('argument is not a matrix; perhaps it is a data frame?')
+	if (is.null(type))
+	{
+		type = typeof(x)
+	}
+  
+  if (type=="integer" | type=="double" | type=="short" | type=="char") 
+  {
+    if (!shared & is.null(backingpath))
+    {
+      y <- big.matrix(nrow=nrow(x), ncol=ncol(x), type=type, init=NULL, 
+				dimnames=dimnames(x), separated=separated)
+    }
+    else 
+    {
+      y = shared.big.matrix(nrow=nrow(x), ncol=ncol(x), type=type, init=NULL, 
+				dimnames=dimnames(x), separated=separated, backingfile=backingfile, 
+				backingpath=backingpath, preserve=preserve)
+    }
     y[1:nrow(x),1:ncol(x)] <- x
-    if (!is.null(colnames(x))) colnames(y) <- colnames(x)
-    if (!is.null(rownames(x))) rownames(y) <- rownames(x)
     junk <- gc() 
   } else stop('that type is not implemented.')
   return(y)
@@ -93,7 +138,9 @@ assign('colnames.bm<-',
         }
       }
       if (!is.null(value) & length(value) != ncol(x))
+      {
         stop("length of 'colnames' not equal to array extent.")
+      }
       .Call("SetColumnNames", x@address, value)
       return(x)
     } else stop('changing column names of a shared object is prohibited.')
@@ -133,89 +180,26 @@ setMethod('nrow', signature(x="big.matrix"),
 setMethod('dim', signature(x="big.matrix"),
   function(x) return(c(nrow(x), ncol(x))))
 
-GetElements.bm = function(x,i,j)
+'[.big.matrix' = function(x, i=NULL, j=NULL)
 {
   if (is.shared(x) && options()$rlock.enabled)
+	{
     lockcols(x, j, 'r')
-  # Get rid of any zero i,j values
-  if (sum(i == 0) > 0) i <- i[i != 0]
-  if (sum(j == 0) > 0) j <- j[j != 0]
-	
-  # Make sure i and j have either all negative or all positive values.
-  if ( (sum(i > 0) > 0 & sum(i < 0) >0) | (sum(j > 0) > 0 & sum(j < 0)) )
+	}
+  retList <- .Call("GetMatrixElements", x@address, j, i)
+
+  dimnames(retList[[1]]) = list( retList[[2]], retList[[3]] )
+  if (sum(dim(retList[[1]])==1) > 0)
   {
-    if (is.shared(x) && options()$rlock.enabled)
-      unlockcols(x,j)
-    stop("only 0's may be mixed with negative subscripts")
-  }
-
-  # Handle the case where i or j are negative indices
-  if (sum(j < 0) > 0) {
-    j = -j
-    j = setdiff( 1:ncol(x), j )
-  }
-  if (sum(i < 0) > 0) {
-    i = -i
-    i = setdiff( 1:nrow(x), i )
-  }
-  if (sum(i > nrow(x)) | sum(j > ncol(x))) stop("Indexes out of range.")
-
-  ret <- .Call("GetMatrixElements", x@address, as.double(j), as.double(i))
-
-  if (length(i) > 1 & length(j) > 1) {
-    dimNames=NULL
-    if (length(.Call("HasRowColNames", x@address)) > 0)
-    {
-      colNames = .Call("GetIndexColNames", x@address, as.double(j))
-      rowNames = .Call("GetIndexRowNames", x@address, as.double(i))
-      dimNames = list( rowNames, colNames )
-    }
-    ret = matrix(ret, nrow=length(i), ncol=length(j), dimnames=dimNames)
+    retList[[1]] = as.vector(retList[[1]])
   }
   if (is.shared(x) && options()$rlock.enabled)
+	{
     unlockcols(x,j)
-  return(ret)
+	}
+  return(retList[[1]])
 }
 
-setMethod("[",
-  signature(x = "big.matrix", drop = "missing"),
-  function(x, i, j) {
-    if (!is.numeric(i) & !is.character(i) & !is.logical(i))
-      stop("row indices must be numeric, logical, or character vectors.")
-    if (!is.numeric(j) & !is.character(j) & !is.logical(j))
-      stop("column indices must be numeric, logical, or character vectors.")
-    if (is.character(i)) 
-      if (is.null(rownames(x))) stop("row names do not exist.")
-      else i <- mmap(i, rownames(x))
-    if (is.character(j))
-      if (is.null(colnames(x))) stop("column names do not exist.")
-      else j <- mmap(j, colnames(x))
-    if (is.logical(i)) {
-      if (length(i) != nrow(x))
-        stop("row vector length must match the number of rows of the matrix.")
-      i <- which(i)
-    }
-    if (is.logical(j)) {
-      if (length(j) != ncol(x))
-        stop(paste("column vector length must match the number of",
-                   "columns of the matrix."))
-      j <- which(j)
-    }
-    return(GetElements.bm(x, i, j))
-  })
-
-setMethod("[",
-  signature(x = "big.matrix", i="missing", drop = "missing"),
-  function(x, j) return(x[1:nrow(x), j]))
-
-setMethod("[",
-  signature(x = "big.matrix", j="missing", drop = "missing"),
-  function(x, i) return(x[i, 1:ncol(x)]))
-
-setMethod("[",
-  signature(x = "big.matrix", i="missing", j="missing", drop = "missing"),
-  function(x) return(GetElements.bm(x, 1:nrow(x), 1:ncol(x))))
-
 SetFunction.bm = function(x, i, j, value)
 {
   if (is.shared(x))
@@ -231,12 +215,13 @@ SetFunction.bm = function(x, i, j, value)
   {
     if (is.shared(x))
     {
-      options(rlock.enabled=TRUE)
       unlockcols(x, j)
+      options(rlock.enabled=TRUE)
     }
     stop('Indices out of range')
   }
 
+  totelts <- as.numeric(length(i)) * as.numeric(length(j))
   # If we are assigning from a matrix, make sure the dimensions agree.
   if (is.matrix(value)){
     if (ncol(value) != length(j) | nrow(value) != length(i))
@@ -249,13 +234,14 @@ SetFunction.bm = function(x, i, j, value)
       }
       stop("Matrix dimensions do not agree with big.matrix instance set size")
     }
-  } else if (length(value) != length(i)*length(j)) {
+  } else if (length(value) != totelts) {
     # Otherwise, make sure we are assigning the correct number of things
     # (rep if necessary)
-    numReps = length(i)*length(j)/length(value)
+    numReps = totelts / length(value)
     if (numReps != round(numReps)) {
       stop("number of items to replace is not a multiple of replacement length")
     }
+    if (totelts > .Machine$integer.max) stop("Too large an assignment for R; bigmemory can fix this, eventually.\n")
     value = rep(value, numReps)
   }
   if (typeof(x) != 'double')
@@ -305,6 +291,15 @@ setMethod('[<-',
                    "columns of the matrix."))
       j <- which(j)
     }
+    if ( options()$bigmemory.typecast.warning &&
+         (typeof(value) == "double") && (typeof(x) != "double") ||
+         (typeof(value) == "integer" && 
+          (typeof(x) != "double" && typeof(x) != "integer")) )
+    {
+      warning(cat("Assignment will down cast from ", typeof(value), " to ",
+        typeof(x), "\nHint: To remove this warning type:  ",
+				"options(big.memory.typecast.warning=FALSE)\n", sep=''))
+    }
     return(SetFunction.bm(x, i, j, value))
   })
 
@@ -339,7 +334,7 @@ setMethod('tail', signature(x="big.matrix"),
 
 setMethod('print', signature(x='big.matrix'), 
   function(x) {
-    if (options()$bm.print.warning==TRUE)
+    if (options()$bigmemory.print.warning==TRUE)
     {
       cat("Warning: This is not advised.  Here is the head of the matrix:\n")
       print(head(x))
@@ -354,11 +349,11 @@ setMethod('print', signature(x='big.matrix'),
   })
 
 ###################################################################
-# x big.matrix	
-# cols	is.numeric or is.character
-# vals	list of scalar or 2-vectors otherwise
+# x big.matrix  
+# cols  is.numeric or is.character
+# vals  list of scalar or 2-vectors otherwise
 # comps could be missing, in which case we'll fill in 'eq' in signature, earlier
-# 	list of comparisons matching dim of associated vals component
+#   list of comparisons matching dim of associated vals component
 
 setGeneric('mwhich', function(x, cols, vals, comps, op = 'AND')
   standardGeneric('mwhich'))
@@ -518,42 +513,40 @@ mwhich.internal <- function(x, cols, vals, comps, op, whichFuncName)
 rm.cols <- function(x, remove.columns)
 {
   if (is.shared(x)) stop("unable to remove columns of a shared matrix.")
-  else {
-    if (length(remove.columns) != unique(length(remove.columns))) 
-      stop("You can only remove a column once.")
-    if (is.character(remove.columns)) 
-      remove.columns= mmap(remove.columns, colnames(x))
-    else if (min(remove.columns)<1 | max(remove.columns)>ncol(x)) 
-      stop("Error: column indices out of range.")
-    remove.columns = sort(remove.columns, decreasing=TRUE)
-    for (rmCol in remove.columns) {
-      .Call('CEraseMatrixCol', x@address, as.double(rmCol))
-      # TODO: For some reason, if I don't put a statement after this call
-      # I get a segfault from R (according to valgrind).  
-      i=1 
-    }
+  if (is.separated(x)) stop("unable to remove columns of a separated matrix.")
+
+  if (length(remove.columns) != unique(length(remove.columns))) 
+    stop("You can only remove a column once.")
+  if (is.character(remove.columns)) 
+    remove.columns= mmap(remove.columns, colnames(x))
+  else if (min(remove.columns)<1 | max(remove.columns)>ncol(x)) 
+    stop("Error: column indices out of range.")
+  remove.columns = sort(remove.columns, decreasing=TRUE)
+  for (rmCol in remove.columns) {
+    .Call('CEraseMatrixCol', x@address, as.double(rmCol))
+    # TODO: For some reason, if I don't put a statement after this call
+    # I get a segfault from R (according to valgrind).  
+    i=1 
   }
 }
 
 add.cols <- function(x, ncol=1, init=0, column.names=NULL)
 {
   if (is.shared(x)) stop("unable to add columns to a shared matrix.")
-  else {
-    if (ncol < 1) stop('You must add at least 1 column.')
-    oldncols = ncol(x)
-    if (is.null(column.names) && !is.null(colnames(x))) 
-    {
-      column.names <- 
-        paste(rep('V', ncol), 
-          (oldncols+1):(ncol(x)+ncol), sep=".")
-    }
-    if (length(unique(c(colnames.bm(x), column.names))) != ncol(x)+ncol)
-      stop("Variable names conflict in add.cols.\n")
-    for (i in 1:ncol) {
-      .Call('CAddMatrixCol', x@address, as.double(init))
-    }
-    if (!is.null(colnames(x))) colnames(x)[-c(1:oldncols)] = column.names
+  if (is.separated(x)) stop("unable to add columns to a separated matrix.")
+
+  if (ncol < 1) stop('You must add at least 1 column.')
+  oldncols = ncol(x)
+  if (is.null(column.names) && !is.null(colnames(x))) {
+    column.names <- paste(rep('V', ncol), 
+                          (oldncols+1):(ncol(x)+ncol), sep=".")
   }
+  if (length(unique(c(colnames.bm(x), column.names))) != ncol(x)+ncol)
+    stop("Variable names conflict in add.cols.\n")
+  for (i in 1:ncol) {
+    .Call('CAddMatrixCol', x@address, as.double(init))
+  }
+  if (!is.null(colnames(x))) colnames(x)[-c(1:oldncols)] = column.names
 }
 
 setMethod('dimnames', signature(x = "big.matrix"),
@@ -568,51 +561,62 @@ setMethod('dimnames<-', signature(x = "big.matrix", value='list'),
 
 hash.mat <- function(x, col)
 {
-  if (colmin(x, col)<1) stop("Error: minimum value in specified column should be 1 or more.")
+  if (colmin(x, col)<1) 
+    stop("Error: minimum value in specified column should be 1 or more.")
   return(matrix(.Call('MatrixHashRanges', x@address, as.double(col)),
                       ncol=2, byrow=TRUE))
 }
 
 read.big.matrix <- function(fileName, sep=',', header=FALSE, row.names=NULL, 
-  col.names=NULL, type=NA, skip=0, shared=FALSE) 
+  col.names=NULL, type=NA, skip=0, separated=FALSE,  shared=FALSE,
+  backingfile=NULL, backingpath=NULL, preserve=TRUE) 
 { stop("Error: You must specify a file name.") }
 
 setMethod('read.big.matrix', signature(fileName='character'),
   function(fileName, sep=',', header=FALSE, row.names=NULL, col.names=NULL,
-           type=NA, skip=0, shared=FALSE)
+           type=NA, skip=0, separated=FALSE, shared=FALSE, 
+           backingfile=NULL, backingpath=NULL, preserve=TRUE)
   {
-
     headerOffset <- as.numeric(header)
     colNames=NULL
     if (header) {
       colNames = unlist(strsplit(
         scan(fileName, what='character', skip=skip, nlines=1, sep="\n", 
-          split=sep, quiet=TRUE)))
+          quiet=TRUE), split=sep))
       if (is.character(col.names)) {
         warning("Using supplied column names and skipping the header row.\n")
         colNames <- col.names
       } else {
-        if (is.logical(col.names) & !col.names)
-          stop("You need to decide whether there are column names.\n")
+        if (!is.null(col.names))
+          stop("You need to decide between column names and a header.\n")
       }
     } else {
       if (is.character(col.names)) colNames <- col.names
     }
+		# Get the first line of data
     firstLineVals <- unlist(strsplit(
       scan(fileName, what='character', skip=(skip+headerOffset), 
            nlines=1, sep="\n", quiet=TRUE), split=sep))
 
-    rowNames <- NULL
-    userowNames <- NULL
+		# See if there are row names (a row name always has a double quote)
+		rowNames=NULL
+		userowNames=FALSE
+		hasQuoteInFirstLine = grep( '"', firstLineVals )
+		hasRowNames = length(hasQuoteInFirstLine) > 0 && hasQuoteInFirstLine > 0
+		if (hasRowNames)
+		{
+			userowNames=TRUE
+		}
+    
     if (length(colNames) == length(firstLineVals)) {
       # No row names, may or may not be column names, but we don't care
       # and nothing more needs to be done at this point.
     }
-    if (!is.null(colNames) && ((length(colNames)+1)==length(firstLineVals))) {
+    if (!is.null(colNames) && (length(colNames)==length(firstLineVals))) {
       # Column and row names both exist
-      if (is.logical(row.names) & !row.names) 
+      if (is.logical(row.names) && !row.names) 
         stop("Error: row names seem to exist in this file.\n")
-      if (is.logical(row.names) & row.names) {
+      if (is.logical(row.names) && row.names) {
         firstLineVals = firstLineVals[-1]
         userowNames <- TRUE
       }
@@ -629,7 +633,7 @@ setMethod('read.big.matrix', signature(fileName='character'),
       }
     }
 
-    numCols <- length(firstLineVals)
+    numCols <- length(firstLineVals) - as.integer(hasRowNames)
     if (is.na(type)) {
       warning('big.matrix type was not specified, going by first line, noting that a choice will be made between double and integer only (not short or char).')
       type <- 'double'
@@ -643,14 +647,24 @@ setMethod('read.big.matrix', signature(fileName='character'),
 
     lineCount <- .Call("CCountLines", fileName) - skip - headerOffset
     numRows <- lineCount
-
-    if (!shared)
+    if (!shared & is.null(backingfile))
     {
       bigMat = big.matrix(nrow=numRows, ncol=numCols, type=type, 
         dimnames=list(rowNames, colNames), init=0)
-    } else {
-      bigMat = shared.big.matrix(nrow=numRows, ncol=numCols, type=type, 
-        dimnames=list(rowNames, colNames), init=0 )
+    } 
+    else 
+    {
+      if (!is.null(backingfile)) 
+      {
+        bigMat = filebacked.big.matrix(nrow=numRows, ncol=numCols, type=type,
+          dimnames=list(rowNames, colNames), init=NULL, separated=separated,
+          backingfile=backingfile, backingpath=backingpath, preserve=preserve)
+      } 
+      else 
+      {
+        bigMat = shared.big.matrix(nrow=numRows, ncol=numCols, type=type, 
+          dimnames=list(rowNames, colNames), init=NULL, separated=separated )
+      }
     }
     ############################################################
     # if userowNames == NULL, then there aren't any in the file.
@@ -658,9 +672,9 @@ setMethod('read.big.matrix', signature(fileName='character'),
     # if userowNames == FALSE, then they exist, but you ignore them.
     .Call('ReadMatrix', fileName, bigMat@address, 
           as.integer(skip+headerOffset), as.integer(numRows), 
-          as.integer(numCols), sep, userowNames)
-    if (is.logical(userowNames) && userowNames == FALSE)
-      rownames(bigMat) = row.names
+          as.integer(numCols), sep, hasRowNames, userowNames)
+#    if (is.logical(userowNames) && userowNames == FALSE)
+#      rownames(bigMat) = row.names
     return(bigMat)
   })
 
@@ -692,21 +706,47 @@ setGeneric('is.shared', function(x) standardGeneric('is.shared'))
 setMethod('is.shared', signature(x='big.matrix'),
   function(x)
   {
-    #func=''
-    #if (x@type == 'integer') func <- "IsIntShared"
-    #if (x@type == 'double') func <- "IsDoubleShared"
-    #if (x@type == 'char') func <- "IsCharShared"
-    #if (x@type == 'short') func <- "IsShortShared"
     return(.Call("IsShared", x@address))
   })
 
 setMethod('is.shared', signature(x='matrix'), function(x) return(FALSE))
 
-deepcopy <- function(x) 
+setGeneric('is.separated', function(x) standardGeneric('is.separated'))
+
+setMethod('is.separated', signature(x='big.matrix'),
+  function(x)
+  {
+    return(.Call("IsSeparated", x@address))
+  })
+
+setMethod('is.separated', signature(x='matrix'), function(x) return(FALSE))
+
+deepcopy <- function(x, type=NULL, separated=NULL, shared=NULL, 
+	backingfile=NULL, backingpath=NULL, preserve=TRUE) 
 {
-  y <- big.matrix(nrow=nrow(x), ncol=ncol(x), type=typeof(x), dimnames=dimnames(x))
-  for (i in 1:ncol(x)) y[,i] <- x[,i]
+	if (nrow(x) > 2^31-1)
+	{
+		stop(paste("Too many rows to copy for bigmemory version 3.2. ",
+			"This will be fixed in the next iteration."))
+	}
+	if (is.null(type))
+	{
+		type = typeof(x)
+	}
+	if (is.null(separated))
+	{
+		separated=is.separated(x)
+	}
+	if (is.null(shared))
+	{
+		shared = is.shared(x)
+	}
+	y = big.matrix( nrow=nrow(x), ncol=ncol(x), type=type, init=NULL,
+		dimnames = dimnames(x), separated=separated, shared=shared,
+		backingfile=backingfile, backingpath=backingpath, preserve=preserve)
+  for (i in 1:ncol(x))
+  {
+    y[,i] <- x[,i]
+  }
   return(y)
 }
-
-
